"""
.. codeauthor:: Nathan Hertz <nhertz@nrao.edu>

Workspaces testing utilities

Put things here if you'd like them to be available to any given test in the entire system

HOW TO USE:
1. In any given pytest-powered test, add this line below the imports:
pytest_plugins = ["testing.utils.conftest"]
2. All fixtures within this file will now be available for use
3. If you want a non-fixture item from here, you'll have to do a normal Python import
"""
import json
from typing import List
from unittest import mock
from unittest.mock import patch, Mock

import pendulum
import pytest
import workspaces

from workspaces.workflow.schema import WorkflowRequest, Workflow, WorkflowRequestFile

from messaging.messenger import Messenger
from messaging.router import Router
from mock_alchemy.mocking import UnifiedAlchemyMagicMock

from workspaces.capability.schema import (
    Capability,
    CapabilityRequest,
    CapabilityVersion,
)
from workspaces.capability.schema_interfaces import (
    CapabilityExecutionIF,
    CapabilityRequestIF,
)
from workspaces.capability.services.capability_engine import CapabilityEngine
from workspaces.capability.services.capability_info import CapabilityInfo
from workspaces.capability.services.capability_service import CapabilityService
from workspaces.capability.services.interfaces import CapabilityEngineIF
from workspaces.workflow.services.workflow_info import WorkflowInfo
from workspaces.workflow.services.workflow_service import WorkflowService


@pytest.fixture(scope="session")
def mock_capability_info() -> CapabilityInfo:
    """
    Pytest fixture providing a mock CapabilityInfo object with a test capability; mocked object can be used
    like a normal CapabilityInfo instance
    TODO: Use UnifiedAlchemyMagicMock's data parameter to add test capabilities

    :return: Mocked CapabilityInfo
    """
    null_capability = Capability(
        name="null",
        steps="prepare-and-run-workflow null\nawait-workflow",
        max_jobs=2,
    )
    test_capability = Capability(name="test", steps="await-qa", max_jobs=777)
    null_capability_query = (
        [mock.call.query(Capability), mock.call.filter_by(name="null")],
        [null_capability],
    )
    test_capability_query = (
        [mock.call.query(Capability), mock.call.filter_by(name="test")],
        [test_capability],
    )
    null_capability_request = CapabilityRequest(
        id=1,
        state="Ready",
        capability=null_capability,
        parameters="-g",
        # a trick here is to ensure that we always have a first version, with the original parameters
        versions=[CapabilityVersion(version_number=1, parameters="-g")],
    )
    null_capability_request_query = (
        [mock.call.query(CapabilityRequest), mock.call.filter_by(id=1)],
        [null_capability_request],
    )
    mocked_capability_info = CapabilityInfo(
        UnifiedAlchemyMagicMock(
            data=[null_capability_query, test_capability_query, null_capability_request_query],
        )
    )

    return mocked_capability_info


@pytest.fixture(scope="session")
def mock_capability_service(
    mock_capability_info: CapabilityInfo,
    mock_workflow_service: WorkflowService,
) -> CapabilityService:
    """
    Pytest fixture providing a mock CapabilityService object with no threaded event listener creation

    :return: Mocked CapabilityService
    """
    with patch("workspaces.capability.services.capability_service.Router"):
        return CapabilityService(mock_capability_info, mock_workflow_service)


@pytest.fixture(scope="session")
def mock_workflow_info() -> WorkflowInfo:
    """
    Pytest fixture providing a mock WorkflowInfo object; mocked object can be used
    like a normal WorkflowInfo instance

    :return: Mocked WorkflowInfo
    """

    null_workflow = Workflow(
        workflow_name="null",
    )

    null_workflow_query = (
        [mock.call.query(Workflow), mock.call.filter_by(name="null")],
        [null_workflow]
    )

    null_request = WorkflowRequest(
        workflow_request_id=1,
        workflow_name=null_workflow.workflow_name,
        argument="-n",
        state="Created",
        results_dir="",
        created_at=pendulum.now(),
        updated_at=pendulum.now()
    )

    null_request_query = (
        [mock.call.query(WorkflowRequest), mock.call.filter_by(id=1)],
        [null_request]
    )

    mock_workflow_info = WorkflowInfo(UnifiedAlchemyMagicMock(data=[null_workflow_query, null_request_query]))
    return mock_workflow_info


@pytest.fixture(scope="session")
def mock_workflow_service(mock_workflow_info: WorkflowInfo) -> WorkflowService:
    """
    Pytest fixture providing a mock WorkflowService object

    :return: Mocked WorkflowService
    """
    with patch("workspaces.workflow.services.workflow_service.Router"):
        return WorkflowService(mock_workflow_info)


@pytest.fixture(scope="session")
def mock_capability_request(
    mock_capability_service: CapabilityService,
) -> List[CapabilityRequestIF]:
    """
    Pytest fixture adding two mock CapabilityRequest objects that can be accessed from the mock_capability_info

    :param mock_capability_service: Mocked CapabilityService
    :return: Mocked CapabilityRequest
    """
    requests = []
    r1 = mock_capability_service.create_request("null", ["-g"])
    r1.id = 0
    requests.append(r1)
    r2 = mock_capability_service.create_request("test")
    r2.id = 1
    requests.append(r2)
    return requests


@pytest.fixture(scope="session")
def mock_capability_execution(
    mock_capability_info: CapabilityInfo, mock_capability_request: CapabilityRequest
) -> CapabilityExecutionIF:
    """
    Pytest fixture adding a mock CapabilityExecution object that can be
    accessed from mock_capability_info

    :param mock_capability_request: Mocked CapabilityRequest
    :param mock_capability_info: Mocked CapabilityInfo
    :return: Mocked CapabilityExecution
    """
    request = mock_capability_request[0]

    execution = mock_capability_info.create_execution(
        mock_capability_info.lookup_capability_request(request)
    )
    return execution


@pytest.fixture(scope="session")
def mock_capability_engine() -> CapabilityEngineIF:
    """
    Pytest fixture adding a mock CapabilityEngine object
    :return: mocked engine
    """

    return CapabilityEngine(mock_capability_info, mock_workflow_service)


class MockMessenger:
    destination = None

    def send_message(self, **message):
        # print(f"sending message: {message}")
        self.destination.receive_message(json.dumps(message))


@pytest.fixture(scope="session")
def test_router() -> Router:
    """
    Pytest fixture providing a Router class for testing

    :return: Test Router
    """
    messenger = MockMessenger()
    router = Router("test", messenger)
    messenger.destination = router
    return router


@pytest.fixture(scope="module")
def test_messenger(test_router: Router) -> Messenger:
    """
    Pytest fixture providing a Messenger class for testing; uses the test router's messenger

    :param test_router: Router class used for testing
    :return: Test Messenger
    """
    return test_router.messenger
